Attribute VB_Name = "parser"
Option Explicit
'
' a local stack with up to 500 variant elements, and the stack index
'
Public stack(500) As Variant
Public stackIndex As Integer
'
' flags for open and close types
'
Enum tagType
    tagOPEN
    tagCLOSE
End Enum
'
' an array of up to 100 legitimate tags & the actual count
'
Public tagStrings(100) As String    ' array of valid tags
Public tagStringsMax As Integer     ' max index on tagStrings
'
' tagsInitialize --- set up the array of legitimate tags
'
' NOTE: this is done dynamically so as to make it easy to add
'       tags anywhere in the array.
'
Public Sub tagsInitialize()
    Dim ix As Integer
    '
    ' the algorithm requires that tags with indexes of 0 and 1
    ' not be used because negative tags are used as error return
    ' flags and 1) negative 0 cannot be distinguished from '
    ' positive 0, and 2) negative 1 is used as a special flag not
    ' associated with a specific tag.
    '
    tagStrings(0) = "not used"
    tagStrings(1) = "not used"
    tagStringsMax = 2
    '
    ' stuff all legit tags into the tag array (NOTE: I KNOW this is NOT
    ' a full list of HTML tags, it's just an arbitrary sub-set that I
    ' stuck in for this sample project)
    '
    tagSetString "a"
    tagSetString "b"
    tagSetString "br"
    tagSetString "body"
    tagSetString "center"
    tagSetString "div"
    tagSetString "font"
    tagSetString "head"
    tagSetString "html"
    tagSetString "img"
    tagSetString "p"
    tagSetString "script"
    tagSetString "table"
    tagSetString "td"
    tagSetString "tr"
    
End Sub
'
' tagSetString --- given a string, make it the next legitimate tag
'
Public Sub tagSetString(stg As String)
    tagStrings(tagStringsMax) = stg
    tagStringsMax = tagStringsMax + 1
End Sub
'
' HTML_line_validate --- for each open tag found, push the
'                        corresponding token, and for each
'                        closing tag found, pop a token and
'                        verify it.
'
'
' set the return value to
'
'   o 0 for a balanced line
'   o the index for a line continuation
'   o error return of the negative of the tag token if
'       there is an unbalanced tag close.  Tags of 0 and +1 are
'       not allowed
'   o -1 for "I don't know what the #&$^% is going on !!!". Note
'       that this special use of -1 is the reason tags of +1 are
'       not allowed
'
Public Function HTML_line_validate(thisLine As String) As Integer
    Dim startIndex As Integer
    Dim stopIndex As Integer
    Dim tempIndex As Integer
    Dim ix As Integer
    Dim tag As String
    Dim thisType As Integer
    '
    ' "forever" loop to parse line.  "forever" is over when you
    ' run out of tags to analyse or when you find an unbalanced
    ' close
    '
    Do While True
        '
        ' look for the start of a tag
        '
        startIndex = InStr(startIndex + 1, thisLine, "<")
        '
        ' if there is no start left in the line, we're done
        ' and if the stack is empty, then as far as we know,
        ' we're balanced, so return a 0 to flag success but
        ' if the stack is not empty, then require a line
        ' continuation
        '
        If startIndex = 0 Then
            If stackIndex > 0 Then
                HTML_line_validate = 1
            Else
                HTML_line_validate = 0
            End If
            Exit Function
        Else
            '
            ' the tag start was found, so look for the tag end
            '
            stopIndex = InStr(startIndex + 1, thisLine, ">")
            '
            ' if there is no tag end, then there is a hanging
            ' tag start, so just return its index for a line
            ' continuation
            '
            If stopIndex = 0 Then
                HTML_line_validate = startIndex
                Exit Function
            Else
                '
                ' there was a tag start, and a tag end, so
                ' analyse the tag
                '
                '
                ' first, let's see if there is a blank in the tag
                '
                tempIndex = InStr(startIndex + 1, thisLine, " ")
                If tempIndex <> 0 And tempIndex < stopIndex Then
                    '
                    ' there's a blank in the tag, so the tag is
                    ' the text from the "<" to the " " exclusive
                    '
                    tag = Mid(thisLine, startIndex + 1, tempIndex - startIndex - 1)
                Else
                    '
                    ' there's no blank in the tag, so the tag is
                    ' the text from the "<" to the ">" exclusive
                    '
                    tag = Mid$(thisLine, startIndex + 1, stopIndex - startIndex - 1)
                End If
                '
                ' OK, now we have the tag, let's make sure it's
                ' valid and then push it (or do error return)
                '
                ' first, let's see if it's an open or close
                ' and if it's a close, then strip off the "/"
                '
                If Mid$(thisLine, startIndex + 1, 1) = "/" Then
                    thisType = tagType.tagCLOSE
                    tag = Right$(tag, Len(tag) - 1)
                Else
                    thisType = tagType.tagOPEN
                End If
                '
                ' set the tag to all lower case for ease of
                ' comparison with element of the tag array
                '
                tag = LCase$(tag)
                '
                ' see if this tag is in the tag array
                '
                For ix = 2 To tagStringsMax
                    '
                    ' if the tag is legit, deal with it
                    '
                    If tag = tagStrings(ix) Then
                        '
                        ' if it's an open, then push the token and
                        ' if it's a close, then pop a token and
                        ' make sure it's the right token
                        '
                        If thisType = tagType.tagOPEN Then
                            stackPush ix
                            Exit For
                        Else
                            If ix = stackPop Then
                                Exit For
                            Else
                                HTML_line_validate = -ix
                                Exit Function
                            End If
                        End If
                    End If
                Next ix
                '
                ' if the tag was not legit, then return a -1
                ' error flag
                '
                If ix = tagStringsMax + 1 Then
                    HTML_line_validate = -1
                    Exit Function
                End If
            End If
        End If
    Loop
End Function
'
' stackPush --- given a variant, push it onto the stack
'
Public Sub stackPush(invar As Variant)
    stack(stackIndex) = invar
    stackIndex = stackIndex + 1
End Sub
'
' stackPop --- pop a variant off of the stack and return it
'
Public Function stackPop() As Variant
    stackIndex = stackIndex - 1
    stackPop = stack(stackIndex)
End Function
'
' stackInitialize --- clear the stack
'
Public Function stackInitialize()
    stackIndex = 0
End Function
